import datetime
from os.path import isfile, join, abspath, dirname, isdir
from os import mkdir, remove
import json

from collections import OrderedDict


ENTRY_TEMPLATE = """
## {version}

**{date}**
{description}

{changes}
"""


def new_entry(d, changes):
    now = datetime.datetime.now()
    return {"description": d, "date": now.isoformat(), "changes": changes}


def to_markdown(data, title=""):
    if not title:
        s = "# change log"
    else:
        s = "# {} change log".format(title)

    entries = []
    for k, d in data.items():
        changes = "\n".join([" - " + c for c in d["changes"]])
        entry = ENTRY_TEMPLATE.format(
            version=k, date=d["date"], description=d["description"], changes=changes
        )
        entries.append(entry)
    s += "\n".join(entries)
    return s


def load(changelogpath):
    if isfile(changelogpath):
        with open(changelogpath, "r") as f:
            text = f.read()
            if text.strip() == "":
                changelog = []
            else:
                changelog = json.loads(text)
    else:
        changelog = {}

    sorted_entries = sorted(
        [(k, v) for k, v in changelog.items()], key=lambda x: x[1]["date"], reverse=True
    )
    return OrderedDict(sorted_entries)


def update(version, description, changes, changelogpath):
    changelog = load(changelogpath)

    if version in changelog:
        entry = changelog[version]
        entry["description"] = description

        now = datetime.datetime.now()
        entry["date"] = now.isoformat()
        unique_changes = []
        for e in entry["changes"] + changes:
            if e not in unique_changes:
                unique_changes.append(e)
        entry["changes"] = unique_changes
    else:
        entry = new_entry(description, changes)
    changelog[version] = entry
    with open(changelogpath, "w") as f:
        json.dump(changelog, f, indent=2)


def save_to_markdown(changelogpath, path, title=""):
    with open(path, "w") as f:
        f.write(to_markdown(load(changelogpath), title=title))


def update_changelog_interactive(version, changelogpath, markdownpath, title=""):
    description = input("Add a description: ")
    change_input = None
    changes = []
    while change_input != "":
        if change_input is not None:
            changes.append(change_input)
        change_input = input("List a change (ENTER to finish): ").strip()
    update(version, description, changes, changelogpath)
    save_to_markdown(changelogpath, markdownpath, title=title)



import os
import toml
from os.path import join, abspath, dirname, isfile, isdir
import fire
from warnings import warn
from .version import __version__, __name__
from .changelog_utils import update_changelog_interactive, save_to_markdown

PYPROJECT = "pyproject.toml"
RED = "\u001b[31m"
RESET = "\u001b[0m"
VERSIONPY = "__version__.py"


def err(msg):
    return RED + msg + RESET


class Pkg(object):
    def __init__(self, directory, filename):
        self.path = join(directory, filename)
        self.directory = directory
        if not isfile(self.path):
            warn(
                err(
                    "Cannot run keats. No {} file found in currenct directory.".format(
                        PYPROJECT
                    )
                )
            )

    def _get(self, key):
        return self.config["tool"]["poetry"].get(key, None)

    @property
    def config(self):
        """
        Return the toml file information.

        :return:
        :rtype:
        """
        return toml.load(self.path)

    def config_info(self):
        toml_info = dict(self.config["tool"]["poetry"])
        pkg_info = {
            "version": toml_info["version"],
            "name": toml_info["name"],
            "title": toml_info["name"],
            "authors": toml_info.get("authors", list()),
            "repo": toml_info.get("repo", None),
            "homepage": toml_info.get("homepage", None),
            "description": toml_info.get("description", None),
        }
        return pkg_info

    def version_py(self):
        return join(self.pkg_path(), VERSIONPY)

    def local_path(self, path):
        return abspath(join(self.directory, path))

    def pkg_path(self):
        return self.local_path(self.package())

    def test_path(self):
        return self.local_path("tests")

    def packages(self):
        """
        Return all listed packaged in the pyproject.toml file. Note that this
        does not consider * files.

        :return: list of package names
        :rtype: list
        """
        if "packages" in self.config["tool"]["poetry"]:
            pkgs = []
            for pkg in self.config["tool"]["poetry"]["packages"]:
                if "from" in pkg:
                    pkgs.append(join(pkg["from"], pkg["include"]))
                else:
                    pkgs.append(pkg["include"])
        else:
            pkgs = [self._get("name")]
        return pkgs

    def package(self):
        """
        Return the estimated main package from the pyproject.toml file.

        :return: version
        :rtype: basestring
        """
        return self.packages()[0]

    def run_cmd(self, *cmd):
        cline = "(cd {}; {})".format(self.directory, " ".join(cmd))
        print(cline)
        return os.system(cline)

    def run_poetry_cmd(self, *cmd):
        return self.run_cmd("poetry run", *cmd)


class Base(object):
    def __init__(self, path):
        self._path = path

    def _cmd(self, *cmd):
        return self._path.run_cmd(*cmd)

    def _get(self, key):
        return self._path.config["tool"]["poetry"].get(key, None)


class Run(Base):
    def format(self):
        """
        Run the format script.

        :return:
        :rtype:
        """
        paths = [self._path.local_path(p) for p in self._path.packages() + ["tests"]]
        self._cmd(" ".join(["black"] + paths))

    def _script(self, name):
        here = abspath(dirname(__file__))
        return join(here, "..", "keat_scripts", name)

    def release(self):
        self._cmd("sh {}".format(self._script("release.sh")))

    def document(self):
        pass


class Version(Base):
    def print(self):
        """Return package version"""
        return self._get("version")

    def bump(self, version=None):
        if version is None:
            self._cmd("poetry version")
        else:
            self._cmd("poetry version {}".format(version))
        self._write()

    def up(self, version=None):
        """
        Update the package version from the pyproject.toml file.

        :param version: if provided, bump the version number (optional)
        :type version: basestring
        :return: the version
        :rtype: basestring
        """
        if version is None:
            self._write()
        else:
            self.bump(version)
        return self._get("version")

    def _write(self, with_confirm=False):
        pkg_info = self._path.config_info()
        path = self._path.version_py()
        if with_confirm:
            ans = input("Write to '{}'?".format(path))
        else:
            ans = ""
        if ans.strip() == "" or ans.strip().lower() == "y":
            with open(path, "w") as f:
                lines = [
                    "# {}\n".format(VERSIONPY),
                    "# autogenerated by Keats=={}\n".format(__version__),
                ]
                for k, v in pkg_info.items():
                    if isinstance(v, str):
                        v = '"{}"'.format(v)
                    lines.append("__{}__ = {}\n".format(k, v))
                f.writelines(lines)
        else:
            print("no files written")


class ChangeLog(Base):
    @property
    def _dir(self):
        d = join(self._path.directory, ".keats")
        if not isdir(d):
            os.mkdir(d)
        return d

    @property
    def _json(self):
        return join(self._dir, "changelog.json")

    @property
    def _markdown(self):
        return join(self._dir, "changelog.md")

    def up(self):
        """Save changelog to a markdown file."""
        save_to_markdown(self._json, self._markdown)

    def clear(self):
        """Clear the changelog files."""
        print(self._json)
        if isfile(self._json):
            os.remove(self._json)
        if isfile(self._markdown):
            os.remove(self._markdown)

    def new(self):
        """Interactively add a changelog entry. Entries are located in the '.keats' folder."""
        update_changelog_interactive(
            self._get("version"), self._json, self._markdown, self._path.package()
        )


class Keats(object):
    def __init__(self, directory=os.getcwd(), filename=PYPROJECT):
        self.pkg = Pkg(directory, filename)

    def info(self):
        """
        Return package information from the toml file.

        :return:
        :rtype:
        """
        return self.pkg.config_info()

    def keats(self):
        """
        Return the keats version number.
        :return:
        :rtype:
        """
        return __name__ + " " + __version__

    @property
    def v(self):
        """Return package version"""
        return self.version.print()

    @property
    def package(self):
        return self.pkg.package()

    @property
    def packages(self):
        return self.pkg.packages()

    @property
    def version(self):
        return Version(self.pkg)

    @property
    def changelog(self):
        return ChangeLog(self.pkg)

    @property
    def run(self):
        return Run(self.pkg)

    def bump(self, version=None):
        self.version.bump(version)
        self.changelog.new()


def main():
    fire.Fire(Keats)



# __version__.py
# autogenerated by Keats==0.1.0
__version__ = "0.1.1"
__name__ = "keats"
__title__ = "keats"
__authors__ = ["jvrana <justin.vrana@gmail.com>"]
__repo__ = None
__homepage__ = None
__description__ = ""



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['keat_scripts', 'keats']

package_data = \
{'': ['*']}

install_requires = \
['fire>=0.1.3,<0.2.0', 'termcolor>=1.1,<2.0', 'toml>=0.10.0,<0.11.0']

entry_points = \
{'console_scripts': ['keats = keats:main']}

setup_kwargs = {
    'name': 'keats',
    'version': '0.1.1',
    'description': '',
    'long_description': None,
    'author': 'jvrana',
    'author_email': 'justin.vrana@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
